@@ -8,19 +8,19 @@ import argparse
 import os
 import pipes
 import re
+import shlex
 import shutil
 import subprocess
 import sys
 from collections.abc import MutableMapping
 from enum import Enum
-from functools import lru_cache
-from functools import partial
+from functools import lru_cache, partial
 from tempfile import TemporaryDirectory
 
 import docker
 from ruamel.yaml import YAML
 
-__version__ = "1.3.0"
+__version__ = "2.0.0"
 
 # name of the environment variable with GitHub token
 GITHUB_TOKEN_KEY = "GITHUB_TOKEN"
@@ -29,6 +29,16 @@ GITHUB_ACTOR_KEY = "GITHUB_ACTOR"
 # name of possible repository keys used in image value
 IMAGE_REPOSITORY_KEYS = {"name", "repository"}
 
+# prefixes added to prerelease versions
+# these do not include the trailing '.'
+# which makes them valid prerelease fields on their own,
+# but they cannot be empty.
+
+# GIT is always added to start our git info
+GIT_PREFIX = "git"
+# this is the _full_ prefix we add to non-prerelease versions
+PRERELEASE_PREFIX = f"0.dev.{GIT_PREFIX}"
+
 # Container builders
 class Builder(Enum):
     DOCKER_BUILD = "docker-build"
@@ -72,6 +82,11 @@ def _check_call(cmd, **kwargs):
 _check_output = partial(_run_cmd, subprocess.check_output)
 
 
+_semver2 = re.compile(
+    r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
+)
+
+
 def _fix_chart_version(version, strict=False):
     """
     Helm 3 requires published chart versions to follow strict semantic versioning.
@@ -86,14 +101,11 @@ def _fix_chart_version(version, strict=False):
     """
     # semver.org provided this regular expression:
     # https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
-    semver2 = re.compile(
-        r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
-    )
 
-    if semver2.fullmatch(version):
+    if _semver2.fullmatch(version):
         return version
 
-    if version.startswith("v") and semver2.fullmatch(version[1:]):
+    if version.startswith("v") and _semver2.fullmatch(version[1:]):
         return version[1:]
 
     message = f"The version in Chart.yaml '{version}' doesn't follow semantic versioning. Helm 3 requires published charts to have strict semantic versions."
@@ -147,23 +159,50 @@ def _get_latest_commit_modifying_path(*paths, **kwargs):
     )
 
 
+@lru_cache()
+def _count_commits(ref):
+    """Return the number of commits for a given ref (branch, commit, etc.)"""
+    return int(
+        _check_output(
+            ["git", "rev-list", "--count", ref],
+            echo=False,
+        )
+        .decode("utf-8")
+        .strip()
+    )
+
+
 def _get_latest_tag(**kwargs):
     """Get the latest tag on a commit in branch or return None."""
+    return _get_latest_tag_and_count(**kwargs)[0]
+
+
+@lru_cache()
+def _get_latest_tag_and_count(ref="HEAD", **kwargs):
+    """Get the latest tag on a commit in branch and the number of commits since then.
+
+    If no tag is found, return (None, total number of commits in the branch).
+    """
+    kwargs.setdefault("echo", False)  # because we handle errors
     try:
         # If the git command output is my-tag-14-g0aed65e,
         # then the return value will become my-tag.
-        return (
+        git_describe = (
             _check_output(
-                ["git", "describe", "--tags", "--long"],
+                ["git", "describe", "--tags", "--long", ref],
                 stderr=subprocess.DEVNULL,  # because we handle errors
                 **kwargs,
             )
             .decode("utf-8")
             .strip()
-            .rsplit("-", maxsplit=2)[0]
         )
+        latest_tag_in_branch, n_commits_since_tag, _g_sha = git_describe.rsplit(
+            "-", maxsplit=2
+        )
+        return latest_tag_in_branch, int(n_commits_since_tag)
     except subprocess.CalledProcessError:
-        return None
+        # no tag found, count all commits
+        return None, _count_commits(ref)
 
 
 def _get_commit_from_tag(tag, **kwargs):
@@ -236,6 +275,23 @@ def _get_image_build_args(image_options, ns):
     return build_args
 
 
+def _get_image_extra_build_command_options(image_options, ns):
+    """
+    Render extraBuildCommandOptions from chartpress.yaml that could be
+    templates, using the provided namespace that contains keys with dynamic
+    values such as LAST_COMMIT or TAG.
+
+    Args:
+    image_options (dict):
+        The dictionary for a given image from chartpress.yaml.
+        Strings in `image_options['extraBuildCommandOptions']` will be rendered
+        and returned.
+    ns (dict): the namespace used when rendering templated arguments
+    """
+    options = image_options.get("extraBuildCommandOptions", [])
+    return [str(option).format(**ns) for option in options]
+
+
 def _get_image_build_context_path(name, options):
     """
     Return the image's contextPath configuration value, or a default value based
@@ -296,6 +352,7 @@ def build_image(
     context_path,
     dockerfile_path=None,
     build_args=None,
+    extra_build_command_options=None,
     *,
     push=False,
     builder=Builder.DOCKER_BUILD,
@@ -321,6 +378,9 @@ def build_image(
         "<context_path>/Dockerfile".
     build_args (dict, optional):
         Dictionary of docker build arguments.
+    extra_build_command_options (list, optional):
+        List of other docker build options to use. Each item should be a string
+        that gets appended to the build command (e.g. "--label=version=0.1.0").
     push (bool, optional):
         Whether to push the image to a registry
     builder (str):
@@ -356,6 +416,8 @@ def build_image(
             cmd.append("--push")
         elif len(platforms) <= 1:
             cmd.append("--load")
+    if extra_build_command_options:
+        cmd.extend(extra_build_command_options)
     _check_call(cmd)
 
     if builder == Builder.DOCKER_BUILD and push:
@@ -439,38 +501,31 @@ def _image_needs_building(image, platforms):
     return _image_needs_pushing(image, platforms)
 
 
-def _get_identifier_from_paths(*paths, long=False):
+def _get_identifier_from_paths(*paths, long=False, base_version=None):
     latest_commit = _get_latest_commit_tagged_or_modifying_paths(*paths, echo=False)
 
-    try:
-        git_describe = (
-            _check_output(
-                ["git", "describe", "--tags", "--long", latest_commit],
-                echo=False,
-                stderr=subprocess.DEVNULL,  # because we handle errors
-            )
-            .decode("utf8")
-            .strip()
-        )
-        latest_tag_in_branch, n_commits, sha = git_describe.rsplit("-", maxsplit=2)
-        # remove "g" prefix output by the git describe command
-        # ref: https://git-scm.com/docs/git-describe#_examples
-        sha = sha[1:]
+    # always use monotonic counter since the beginning of the branch
+    # avoids counter resets when using explicit base_version
+    n_commits = _count_commits(latest_commit)
 
-        return _get_identifier_from_parts(latest_tag_in_branch, n_commits, sha, long)
-    except subprocess.CalledProcessError:
-        # no tags on branch, so assume 0.0.1 and
-        # calculate n_commits from latest_commit
-        n_commits = (
-            _check_output(
-                ["git", "rev-list", "--count", latest_commit],
-                echo=False,
-            )
-            .decode("utf-8")
-            .strip()
-        )
+    latest_tag_in_branch, n_commits_since_tag = _get_latest_tag_and_count(latest_commit)
+    if latest_tag_in_branch:
+        if n_commits_since_tag == 0:
+            # don't use baseVersion config for development versions
+            # when we are exactly on a tag
+            base_version = latest_tag_in_branch
+            if not long:
+                # set n_commits=0 to ensure exact tag is used without suffix
+                # in _get_identifier_from_parts
+                n_commits = 0
 
-        return _get_identifier_from_parts("0.0.1", n_commits, latest_commit, long)
+        if base_version is None:
+            base_version = latest_tag_in_branch
+
+    if base_version is None:
+        base_version = "0.0.1-0.dev"
+
+    return _get_identifier_from_parts(base_version, n_commits, latest_commit, long)
 
 
 def _get_identifier_from_parts(tag, n_commits, commit, long):
@@ -481,24 +536,37 @@ def _get_identifier_from_parts(tag, n_commits, commit, long):
     This function should provide valid Helm chart versions, which means they
     need to be valid SemVer 2 version strings. It also needs to return valid
     image tags, which means they need to not contain `+` signs either. We prefix
-    with n and h to ensure we don't get a numerical hash starting with 0 because
+    with h to ensure we don't get a numerical hash starting with 0 because
     those are invalid SemVer 2 versions.
 
+    If tag is already a prelease, we append fields so we sort later than the prerelease.
+    If the tag is a stable release, we start the prerelease with `-0.dev.` to ensure we sort lower
+    than labeled prereleases such as 'alpha'.
+
+    Typical tags in descending order:
+
+    - 0.1.2
+    - 0.1.2-alpha.2
+    - 0.1.2-alpha.1.git.5.habc
+    - 0.1.2-alpha.1
+    - 0.1.2-0.dev.git.3.habc
+
     Example:
         tag="0.1.2", n_commits="5", commit="asdf1234", long=True,
-        should return "0.1.2-n005.hasdf1234".
+        should return "0.1.2-0.dev.git.5.hasdf1234".
     """
     n_commits = int(n_commits)
 
     if n_commits > 0 or long:
         if "-" in tag:
-            # append a pre-release tag, with a . separator
-            # 0.1.2-alpha.1 -> 0.1.2-alpha.1.n.h
-            return f"{tag}.n{n_commits:03d}.h{commit}"
+            # add fields to an existing prerelease tag
+            # 0.1.2-alpha.1 -> 0.1.2-alpha.1.git.n.hsha
+            pre = f".{GIT_PREFIX}"
         else:
-            # append a release tag, with a - separator
-            # 0.1.2 -> 0.1.2-n.h
-            return f"{tag}-n{n_commits:03d}.h{commit}"
+            # create a new '-0.dev.' prerelease tag
+            # 0.1.2 -> 0.1.2-0.dev.git.5.hsha
+            pre = f"-{PRERELEASE_PREFIX}"
+        return f"{tag}{pre}.{n_commits}.h{commit}"
     else:
         return f"{tag}"
 
@@ -515,6 +583,7 @@ def build_images(
     *,
     builder=Builder.DOCKER_BUILD,
     platforms=None,
+    base_version=None,
 ):
     """Build a collection of docker images
 
@@ -542,16 +611,19 @@ def build_images(
 
         Example 1:
         - long=False: 0.9.0
-        - long=True:  0.9.0-n000.hasdf1234
+        - long=True:  0.9.0-0.dev.git.0.hasdf1234
 
         Example 2:
-        - long=False: 0.9.0-n004.hsdfg2345
-        - long=True:  0.9.0-n004.hsdfg2345
+        - long=False: 0.9.0-0.dev.git.4.hsdfg2345
+        - long=True:  0.9.0-0.dev.git.4.hsdfg2345
 
     builder (str):
         The container build engine.
     platforms (list[str]):
         List of platforms to build for if not the same as the Docker host.
+    base_version (str):
+        The base version string (before '.git'), used when useChartVersion is True
+        instead of the tag found via `git describe`.
     """
     values_file_modifications = {}
     for name, options in images.items():
@@ -560,7 +632,9 @@ def build_images(
         all_image_paths = _get_all_image_paths(name, options)
 
         if tag is None:
-            image_tag = _get_identifier_from_paths(*all_image_paths, long=long)
+            image_tag = _get_identifier_from_paths(
+                *all_image_paths, long=long, base_version=base_version
+            )
         else:
             image_tag = tag
 
@@ -592,18 +666,19 @@ def build_images(
 
         # build image and optionally push image
         if force_build or _image_needs_building(image_spec, platforms):
+            expansion_namespace = {
+                "LAST_COMMIT": _get_latest_commit_tagged_or_modifying_paths(
+                    *all_image_paths, echo=False
+                ),
+                "TAG": image_tag,
+            }
             build_image(
                 image_spec,
                 _get_image_build_context_path(name, options),
                 dockerfile_path=_get_image_dockerfile_path(name, options),
-                build_args=_get_image_build_args(
-                    options,
-                    {
-                        "LAST_COMMIT": _get_latest_commit_tagged_or_modifying_paths(
-                            *all_image_paths, echo=False
-                        ),
-                        "TAG": image_tag,
-                    },
+                build_args=_get_image_build_args(options, expansion_namespace),
+                extra_build_command_options=_get_image_extra_build_command_options(
+                    options, expansion_namespace
                 ),
                 push=push or force_push,
                 builder=builder,
@@ -702,23 +777,45 @@ def _update_values_file_with_modifications(name, modifications):
         yaml.dump(values, f)
 
 
+_suffix_version_pat = re.compile(r"(.*)\.git\.\d+\.h[a-f0-9]+$")
+
+
+def _trim_version_suffix(version):
+    """Trim trailing .git... suffix from a version
+
+    Turns 1.0.0-0.dev.git.5.habc back into 1.0.0-0.dev
+    """
+    m = _suffix_version_pat.match(version)
+    if m:
+        # matches, strip suffix
+        return m.group(1)
+    return version
+
+
 def build_chart(
-    name, version=None, paths=None, long=False, reset=False, strict_version=False
+    name,
+    version=None,
+    paths=None,
+    long=False,
+    strict_version=False,
+    base_version=None,
 ):
     """
     Update Chart.yaml's version, using specified version or by constructing one.
 
     Chart versions are constructed using:
-        a) the latest commit that is tagged,
-        b) the latest commit that modified provided paths
+        a) a base version, derived from either Chart.yaml's version field or the latest git tag on branch
+        b) the latest commit that was tagged on the current branch (n)
+        c) the latest commit that modified provided paths (hash)
+
 
     Example versions constructed:
+        - 0.9.0-0.dev.git.2.hdfgh3456
         - 0.9.0-alpha.1
-        - 0.9.0-alpha.1.n000.hasdf1234 (--long)
-        - 0.9.0-alpha.1.n005.hsdfg2345
-        - 0.9.0-alpha.1.n005.hsdfg2345 (--long)
+        - 0.9.0-alpha.1.git.0.hasdf1234 (--long)
+        - 0.9.0-alpha.1.git.5.hsdfg2345
+        - 0.9.0-alpha.1.git.5.hsdfg2345 (--long)
         - 0.9.0
-        - 0.9.0-n002.hdfgh3456
     """
     # read Chart.yaml
     chart_file = os.path.join(name, "Chart.yaml")
@@ -727,16 +824,21 @@ def build_chart(
 
     # decide a version string
     if version is None:
-        version = _get_identifier_from_paths(*paths, long=long)
-    if not reset:
-        version = _fix_chart_version(version, strict=strict_version)
+        # derive the full version, with possible '.git...'
+        version = _get_identifier_from_paths(
+            *paths, long=long, base_version=base_version
+        )
+
+    version = _fix_chart_version(version, strict=strict_version)
 
     # update Chart.yaml
-    if chart["version"] != version:
+    if chart["version"] == version:
+        _log(f"Leaving {chart_file} version unchanged: {version}")
+    else:
         _log(f"Updating {chart_file}: version: {version}")
         chart["version"] = version
-    with open(chart_file, "w") as f:
-        yaml.dump(chart, f)
+        with open(chart_file, "w") as f:
+            yaml.dump(chart, f)
 
     # return version
     return version
@@ -800,15 +902,17 @@ def publish_pages(
     _check_call(["git", "checkout", "gh-pages"], cwd=checkout_dir, echo=True)
 
     # check if a chart with the same name and version has already been published. If
-    # there is, the behaviour depends on `-force-publish-chart`
+    # there is, the behaviour depends on `--force-publish-chart`
     # and chart_version and make a decision based on the --force-publish-chart
     # flag if that is the case, but always log what's done
     if os.path.isfile(os.path.join(checkout_dir, "index.yaml")):
         with open(os.path.join(checkout_dir, "index.yaml")) as f:
             chart_repo_index = yaml.load(f)
-            published_charts = chart_repo_index["entries"].get(chart_name)
+            published_charts = chart_repo_index["entries"].get(chart_name, [])
 
-        if any(c["version"] == chart_version for c in published_charts):
+        if published_charts and any(
+            c["version"] == chart_version for c in published_charts
+        ):
             if force:
                 _log(
                     f"Chart of version {chart_version} already exists, overwriting it."
@@ -863,6 +967,60 @@ def publish_pages(
     _check_call(["git", "push", "origin", "gh-pages"], cwd=checkout_dir)
 
 
+def _check_base_version(base_version):
+    """Verify that a baseVersion config is valid
+
+    If specified, base version needs to:
+
+    1. be a valid semver prerelease
+    2. sort after the latest tag on the branch
+    """
+
+    # check valid value (baseVersion must be semver prerelease)
+    match = _semver2.fullmatch(base_version)
+    if not match:
+        raise ValueError(
+            f"baseVersion: {base_version} must be a valid semver prerelease (e.g. 1.2.3-0.dev), but doesn't appear to be valid."
+        )
+    base_version_groups = match.groupdict()
+    if not base_version_groups["prerelease"]:
+        raise ValueError(
+            f"baseVersion: {base_version} must be a valid semver prerelease (e.g. 1.2.3-0.dev), but is not a prerelease."
+        )
+
+    def _version_number(groups):
+        """Return comparable semver"""
+
+        return (
+            int(groups["major"]),
+            int(groups["minor"]),
+            int(groups["patch"]),
+        )
+
+    # check ordering with latest tag
+    # do not check on a tagged commit
+    tag, count = _get_latest_tag_and_count()
+    if tag and count:
+        tag_match = _semver2.fullmatch(tag.lstrip("v"))
+        sort_error = f"baseVersion {base_version} is not greater than latest tag {tag}. Please update baseVersion config in chartpress.yaml."
+        if tag_match:
+            base_version_number = _version_number(base_version_groups)
+            tag_version_number = _version_number(tag_match.groupdict())
+            if base_version_number < tag_version_number:
+                raise ValueError(sort_error)
+            elif base_version_number == tag_version_number:
+                # numbers equal, need to check prerelease
+                if tag_match["prerelease"]:
+                    # If we want to be pedantic about inter-prerelease ordering (alpha before beta, etc.),
+                    # that would go here. We should import a full semver implementation for that, though.
+                    pass
+                else:
+                    raise ValueError(sort_error)
+        else:
+            # tag not semver. ignore? Not really our problem.
+            _log(f"Latest tag {tag} does not appear to be a semver version")
+
+
 class ActionStoreDeprecated(argparse.Action):
     """Used with argparse as a deprecation action."""
 
@@ -881,7 +1039,7 @@ class ActionAppendDeprecated(argparse.Action):
         getattr(namespace, self.dest).append(values)
 
 
-def main(args=None):
+def main(argv=None):
     """Run chartpress"""
     argparser = argparse.ArgumentParser(description=__doc__)
 
@@ -970,24 +1128,33 @@ def main(args=None):
     )
     argparser.add_argument(
         "--version",
-        action="store_true",
-        help="Print current chartpress version and exit.",
+        action="version",
+        version=f"chartpress version {__version__}",
     )
 
-    args = argparser.parse_args(args)
+    args = argparser.parse_args(argv)
     if args.builder == Builder.DOCKER_BUILD and args.platform:
         argparser.error(f"--platform is not supported with {Builder.DOCKER_BUILD}")
 
+    if args.reset:
+        # reset conflicts with everything
+        # this could probably be clearer by using subparsers
+        argv = list(argv or sys.argv[1:])
+        if len(argv) > 1:
+            argv = list(argv)
+            argv.remove("--reset")
+            extra_args = " ".join(shlex.quote(arg) for arg in argv if arg != "--reset")
+            argparser.error(
+                f"`chartpress --reset` takes no additional arguments: {extra_args}"
+            )
+
     # allow simple checks for whether publish will happen
     if args.force_publish_chart:
         args.publish_chart = True
 
-    if args.version:
-        print(f"chartpress version {__version__}")
-        return
-
-    if args.list_images:
+    if args.list_images or args.reset:
         args.no_build = True
+        args.publish_chart = False
 
     with open("chartpress.yaml") as f:
         config = yaml.load(f)
@@ -1001,34 +1168,52 @@ def main(args=None):
     #   - push chart (--publish-chart, --extra-message)
     for chart in config["charts"]:
         forced_version = None
+        base_version = None
+
         if args.tag:
+            # tag specified, use that version
             forced_version = args.tag
-        if args.reset:
+        elif args.reset:
             forced_version = chart.get("resetVersion", "0.0.1-set.by.chartpress")
 
+        if not args.tag:
+            # Load base_version config when building a dev tag
+            # (i.e. not a tagged commit or explicit tag).
+            # add the check here so `--reset` will fail with an invalid baseVersion
+            # (e.g. forgetting to update after release)
+            base_version = chart.get("baseVersion", None)
+            if base_version:
+                _check_base_version(base_version)
+
         if not args.list_images:
             # update Chart.yaml with a version
             chart_version = build_chart(
                 chart["name"],
                 paths=_get_all_chart_paths(chart),
                 version=forced_version,
+                base_version=base_version,
                 long=args.long,
-                reset=args.reset,
                 strict_version=args.publish_chart,
             )
 
         if "images" in chart:
+            # set common image tag if `--tag` specified _or_ resetting
+            common_image_tag = None
+            if args.tag:
+                common_image_tag = args.tag
+            elif args.reset:
+                common_image_tag = chart.get("resetTag", "set-by-chartpress")
+
             # build images
             values_file_modifications = build_images(
                 prefix=args.image_prefix or chart.get("imagePrefix", ""),
                 images=chart["images"],
-                tag=args.tag
-                if not args.reset
-                else chart.get("resetTag", "set-by-chartpress"),
+                tag=common_image_tag,
                 push=args.push,
                 force_push=args.force_push,
                 force_build=args.force_build,
                 skip_build=args.no_build or args.reset,
+                base_version=base_version,
                 long=args.long,
                 builder=args.builder,
                 platforms=args.platform,
